# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import ./types

when defined(nimdoc):
    type TYPES* = DataType | Attribute | AttributeList

import std/strformat
import ./util

func size*(dtype: DataType): int16 =
    assert dtype != Unknown, "Can't use unknown type"
    case dtype:
    of Float, Int, UInt: 4
    of HalfFloat, Short, UShort: 2
    of Byte, UByte: 1
    of Unknown: 0

func stride*(layout: AttributeList, align: int32 = 4): int32 =
    # Stride of an attribute list, aligned to a byte boundary (default 4)
    for a in layout:
        result = max(result, a.offset + a.dtype.size * a.count)
    result = align(result, align)

proc add*(layout: var AttributeList, attr: Attribute) =
    # Adds an attribute to a seq setting the approriate location and offset
    # (aligned to a 4 byte boundary)
    var attr = attr
    if layout.len != 0:
        let last = layout[layout.high]
        if attr.offset == 0:
            attr.offset = align4 last.offset + last.dtype.size*last.count
        if attr.location == 0:
            attr.location = last.location + 1
    # layout.add y=attr
    layout.setLen layout.len + 1
    layout[^1] = attr

proc contains*(layout: AttributeList, name: string): bool =
    for attr in layout:
        if attr.name == name:
            return true

proc `[]`*(layout: AttributeList, name: string): Attribute =
    for attr in layout:
        if attr.name == name:
            return attr
    raise newException(KeyError, &"key {name} not found")
