# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import vmath except Quat

type KeyCode* = enum
    KeyUnknown = - 1, KeySpace = 32, KeyApostrophe = 39, KeyComma = 44, KeyMinus,
    KeyPeriod, KeySlash, Key0, Key1, Key2, Key3, Key4, Key5, Key6, Key7, Key8,
    Key9, KeySemicolon = 59, KeyEqual = 61, KeyA = 65, KeyB, KeyC, KeyD, KeyE,
    KeyF, KeyG, KeyH, KeyI, KeyJ, KeyK, KeyL, KeyM, KeyN, KeyO, KeyP, KeyQ, KeyR,
    KeyS, KeyT, KeyU, KeyV, KeyW, KeyX, KeyY, KeyZ, KeyLeftBracket, KeyBackslash,
    KeyRightBracket, KeyGraveAccent = 96, KeyWorld1 = 161, KeyWorld2,
    KeyEscape = 256, KeyEnter, KeyTab, KeyBackspace, KeyInsert, KeyDelete,
    KeyRight, KeyLeft, KeyDown, KeyUp, KeyPageUp, KeyPageDown, KeyHome, KeyEnd,
    KeyCapsLock = 280, KeyScrollLock, KeyNumLock, KeyPrintScreen, KeyPause,
    KeyF1 = 290, KeyF2, KeyF3, KeyF4, KeyF5, KeyF6, KeyF7, KeyF8, KeyF9, KeyF10,
    KeyF11, KeyF12, KeyF13, KeyF14, KeyF15, KeyF16, KeyF17, KeyF18, KeyF19,
    KeyF20, KeyF21, KeyF22, KeyF23, KeyF24, KeyF25, KeyKp0 = 320, KeyKp1, KeyKp2,
    KeyKp3, KeyKp4, KeyKp5, KeyKp6, KeyKp7, KeyKp8, KeyKp9, KeyKpDecimal,
    KeyKpDivide, KeyKpMultiply, KeyKpSubtract, KeyKpAdd, KeyKpEnter, KeyKpEqual,
    KeyLeftShift = 340, KeyLeftControl, KeyLeftAlt, KeyLeftSuper, KeyRightShift,
    KeyRightControl, KeyRightAlt, KeyRightSuper, KeyMenu,
    KeyNavigationBack = 400, # Android
    KeyMediaSelect, KeyMediaPlayPause, # tvOS

type MouseButton* = enum
    LeftButton
    RightButton
    MiddleButton
    BackButton
    ForwardButton
    MouseButton6
    MouseButton7
    MouseButton8

type KeyEvent* = object
    pressed*, repeat*: bool
    shiftKey*, ctrlKey*, altKey*, metaKey*: bool
    key*: KeyCode

type MouseButtonEvent* = object
    pressed*: bool
    button*: MouseButton
    shiftKey*, ctrlKey*, altKey*, metaKey*: bool
    position*: Vec2

type MouseMoveEvent* = object
    left*, middle*, right*: bool
    # shiftKey*, ctrlKey*, altKey*, metaKey*: bool
    position*: Vec2
    movement*: Vec2

type MouseWheelEvent* = object
    # shiftKey*, ctrlKey*, altKey*, metaKey*: bool
    movement*: Vec2

