
import std/tables

const glsl_functions* = {
    "rgb_to_hsv": """vec4 rgb_to_hsv(vec4 rgb)
    {
        float cmax, cmin, h, s, v, cdelta;
        vec3 c;

        cmax = max(rgb[0], max(rgb[1], rgb[2]));
        cmin = min(rgb[0], min(rgb[1], rgb[2]));
        cdelta = cmax - cmin;

        v = cmax;
        if (cmax != 0.0)
            s = cdelta / cmax;
        else {
            s = 0.0;
            h = 0.0;
        }

        if (s == 0.0) {
            h = 0.0;
        }
        else {
            c = (vec3(cmax, cmax, cmax) - rgb.xyz) / cdelta;

            if (rgb.x == cmax) h = c[2] - c[1];
            else if (rgb.y == cmax) h = 2.0 + c[0] -  c[2];
            else h = 4.0 + c[1] - c[0];

            h /= 6.0;

            if (h < 0.0)
                h += 1.0;
        }

        return vec4(h, s, v, rgb.w);
    }""",
    "hsv_to_rgb": """vec4 hsv_to_rgb(vec4 hsv)
    {
        float i, f, p, q, t, h, s, v;
        vec3 rgb;

        h = hsv[0];
        s = hsv[1];
        v = hsv[2];

        if (s == 0.0) {
            rgb = vec3(v, v, v);
        }
        else {
            if (h == 1.0)
                h = 0.0;

            h *= 6.0;
            i = floor(h);
            f = h - i;
            rgb = vec3(f, f, f);
            p = v * (1.0 - s);
            q = v * (1.0 - (s * f));
            t = v * (1.0 - (s * (1.0 - f)));

            if (i == 0.0) rgb = vec3(v, t, p);
            else if (i == 1.0) rgb = vec3(q, v, p);
            else if (i == 2.0) rgb = vec3(p, v, t);
            else if (i == 3.0) rgb = vec3(p, q, v);
            else if (i == 4.0) rgb = vec3(t, p, v);
            else rgb = vec3(v, p, q);
        }

        return vec4(rgb, hsv.w);
    }""",
    "mix_overlay": """vec4 mix_overlay(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;
        outc = a;

        if (outc.r < 0.5)
            outc.r *= facm + 2.0 * fac * b.r;
        else
            outc.r = 1.0 - (facm + 2.0 * fac * (1.0 - b.r)) * (1.0 - outc.r);

        if (outc.g < 0.5)
            outc.g *= facm + 2.0 * fac * b.g;
        else
            outc.g = 1.0 - (facm + 2.0 * fac * (1.0 - b.g)) * (1.0 - outc.g);

        if (outc.b < 0.5)
            outc.b *= facm + 2.0 * fac * b.b;
        else
            outc.b = 1.0 - (facm + 2.0 * fac * (1.0 - b.b)) * (1.0 - outc.b);
        return outc;
    }""",
    "mix_dodge": """vec4 mix_dodge(vec4 a, vec4 b, float fac, out vec4 outc){
        outc = a;

        if (outc.r != 0.0) {
            float tmp = 1.0 - fac * b.r;
            if (tmp <= 0.0)
                outc.r = 1.0;
            else if ((tmp = outc.r / tmp) > 1.0)
                outc.r = 1.0;
            else
                outc.r = tmp;
        }
        if (outc.g != 0.0) {
            float tmp = 1.0 - fac * b.g;
            if (tmp <= 0.0)
                outc.g = 1.0;
            else if ((tmp = outc.g / tmp) > 1.0)
                outc.g = 1.0;
            else
                outc.g = tmp;
        }
        if (outc.b != 0.0) {
            float tmp = 1.0 - fac * b.b;
            if (tmp <= 0.0)
                outc.b = 1.0;
            else if ((tmp = outc.b / tmp) > 1.0)
                outc.b = 1.0;
            else
                outc.b = tmp;
        }
        return outc;
    }""",
    "mix_burn": """vec4 mix_burn(vec4 a, vec4 b, float fac, out vec4 outc){
        float tmp, facm = 1.0 - fac;
        outc = a;

        tmp = facm + fac * b.r;
        if (tmp <= 0.0)
            outc.r = 0.0;
        else if ((tmp = (1.0 - (1.0 - outc.r) / tmp)) < 0.0)
            outc.r = 0.0;
        else if (tmp > 1.0)
            outc.r = 1.0;
        else
            outc.r = tmp;

        tmp = facm + fac * b.g;
        if (tmp <= 0.0)
            outc.g = 0.0;
        else if ((tmp = (1.0 - (1.0 - outc.g) / tmp)) < 0.0)
            outc.g = 0.0;
        else if (tmp > 1.0)
            outc.g = 1.0;
        else
            outc.g = tmp;

        tmp = facm + fac * b.b;
        if (tmp <= 0.0)
            outc.b = 0.0;
        else if ((tmp = (1.0 - (1.0 - outc.b) / tmp)) < 0.0)
            outc.b = 0.0;
        else if (tmp > 1.0)
            outc.b = 1.0;
        else
            outc.b = tmp;
        return outc;
    }""",
    "mix_hue": """vec4 mix_hue(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;

        outc = a;

        vec4 hsv, hsv2, tmp;
        hsv2 = rgb_to_hsv(b);

        if (hsv2.y != 0.0) {
            hsv = rgb_to_hsv(outc);
            hsv.x = hsv2.x;
            tmp = hsv_to_rgb(hsv);

            outc = mix(outc, tmp, fac);
            outc.a = a.a;
        }
        return outc;
    }""",
    "mix_saturation": """vec4 mix_saturation(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;

        outc = a;

        vec4 hsv, hsv2;
        hsv = rgb_to_hsv(outc);

        if (hsv.y != 0.0) {
            hsv2 = rgb_to_hsv(b);

            hsv.y = facm * hsv.y + fac * hsv2.y;
            outc = hsv_to_rgb(hsv);
        }
        return outc;
    }""",
    "mix_value": """vec4 mix_value(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;

        vec4 hsv, hsv2;
        hsv = rgb_to_hsv(a);
        hsv2 = rgb_to_hsv(b);

        hsv.z = facm * hsv.z + fac * hsv2.z;
        outc = hsv_to_rgb(hsv);
        return outc;
    }""",
    "mix_color": """vec4 mix_color(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;

        outc = a;

        vec4 hsv, hsv2, tmp;
        hsv2 = rgb_to_hsv(b);

        if (hsv2.y != 0.0) {
            hsv = rgb_to_hsv(outc);
            hsv.x = hsv2.x;
            hsv.y = hsv2.y;
            tmp = hsv_to_rgb(hsv);

            outc = mix(outc, tmp, fac);
            outc.a = a.a;
        }
        return outc;
    }""",
    "mix_soft_light": """vec4 mix_soft_light(vec4 a, vec4 b, float fac, out vec4 outc){
        float facm = 1.0 - fac;

        vec4 one = vec4(1.0);
        vec4 scr = one - (one - b) * (one - a);
        outc = facm * a + fac * ((one - a) * b * a + a * scr);
        return outc;
    }""",
    "blackbody": """vec4 blackbody(float T)
    {
        float u = ( 0.860117757 + 1.54118254e-4 * T + 1.28641212e-7 * T*T ) / ( 1.0 + 8.42420235e-4 * T + 7.08145163e-7 * T*T );
        float v = ( 0.317398726 + 4.22806245e-5 * T + 4.20481691e-8 * T*T ) / ( 1.0 - 2.89741816e-5 * T + 1.61456053e-7 * T*T );

        float x = 3.0*u / ( 2.0*u - 8.0*v + 4.0 );
        float y = 2.0*v / ( 2.0*u - 8.0*v + 4.0 );
        float z = 1.0 - x - y;

        float Y = 1.0;
        float X = Y/y * x;
        float Z = Y/y * z;

        mat3 XYZtoRGB = mat3(
            3.2404542, -0.9692660,  0.0556434,
            -1.5371385, 1.8760108, -0.2040259,
            -0.4985314,  0.0415560, 1.0572252
        );

        return vec4(( XYZtoRGB * vec3( X, Y, Z ) ), 1.0);
    }""",
    "map_range_linear": """float map_range_linear(float v, float min0, float max0, float min1, float max1, float steps){
        float f = (v-min0)/(max0-min0);
        return mix(min1, max1, f);
    }
    vec3 map_range_linear(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = (v-min0)/(max0-min0);
        return mix(min1, max1, f);
    }
    """,
    "map_range_linear_clamp": """float map_range_linear_clamp(float v, float min0, float max0, float min1, float max1, float steps){
        float f = (v-min0)/(max0-min0);
        return mix(min1, max1, clamp(f, 0.0, 1.0));
    }
    vec3 map_range_linear_clamp(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = (v-min0)/(max0-min0);
        return mix(min1, max1, clamp(f, 0.0, 1.0));
    }
    """,
    "map_range_stepped": """float map_range_stepped(float v, float min0, float max0, float min1, float max1, float steps){
        float f = (v-min0)/(max0-min0);
        f = floor(f*(steps+1.0))/(steps);
        return mix(min1, max1, f);
    }
    vec3 map_range_stepped(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = (v-min0)/(max0-min0);
        f = floor(f*(steps+1.0))/(steps);
        return mix(min1, max1, f);
    }
    """,
    "map_range_stepped_clamp": """float map_range_stepped_clamp(float v, float min0, float max0, float min1, float max1, float steps){
        float f = (v-min0)/(max0-min0);
        f = floor(f*(steps+1.0))/(steps);
        return mix(min1, max1, clamp(f, 0.0, 1.0));
    }
    vec3 map_range_stepped_clamp(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = (v-min0)/(max0-min0);
        f = floor(f*(steps+1.0))/(steps);
        return mix(min1, max1, clamp(f, 0.0, 1.0));
    }
    """,
    "map_range_smooth": """float map_range_smooth(float v, float min0, float max0, float min1, float max1, float steps){
        float f = clamp((v-min0)/(max0-min0), 0.0, 1.0);
        f = f * f * (3.0 - 2.0 * f);
        return mix(min1, max1, f);
    }
    vec3 map_range_smooth(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = clamp((v-min0)/(max0-min0), 0.0, 1.0);
        f = f * f * (3.0 - 2.0 * f);
        return mix(min1, max1, f);
    }
    """,
    "map_range_smoother": """float map_range_smoother(float v, float min0, float max0, float min1, float max1, float steps){
        float f = clamp((v-min0)/(max0-min0), 0.0, 1.0);
        f = f * f * f * (f * (f * 6.0 - 15.0) + 10.0);
        return mix(min1, max1, f);
    }
    vec3 map_range_smoother(vec3 v, vec3 min0, vec3 max0, vec3 min1, vec3 max1, vec3 steps){
        vec3 f = clamp((v-min0)/(max0-min0), 0.0, 1.0);
        f = f * f * f * (f * (f * 6.0 - 15.0) + 10.0);
        return mix(min1, max1, f);
    }
    """,
    "math_smoothmin": """float math_smoothmin(float a, float b, float c){
        if (c != 0.0) {
            float h = max(c - abs(a - b), 0.0) / c;
            return min(a, b) - h * h * h * c * (1.0 / 6.0);
        }
        else {
            return min(a, b);
        }
    }""",
    "wrap": """float wrap(float a, float b, float c){
        float range = b - c;
        return (range != 0.0) ? a - (range * floor((a - c) / range)) : c;
    }""",
    "project": """vec3 project(vec3 a, vec3 b){
        float len_sq = dot(b, b);
        return (len_sq != 0.0) ? (dot(a, b) / len_sq) * b : vec3(0.0);
    }""",
    "modulo": """vec3 modulo(vec3 a, vec3 b){
        return vec3(
            sign(a.r) * ((b.r != 0.0) ? mod(abs(a.r),abs(b.r)) : 0.0),
            sign(a.g) * ((b.g != 0.0) ? mod(abs(a.g),abs(b.g)) : 0.0),
            sign(a.b) * ((b.b != 0.0) ? mod(abs(a.b),abs(b.b)) : 0.0) 
        );
    }""",
    "euler_xyz_to_mat3": """mat3 euler_xyz_to_mat3(vec3 euler){
        float cx = cos(euler.x);
        float cy = cos(euler.y);
        float cz = cos(euler.z);
        float sx = sin(euler.x);
        float sy = sin(euler.y);
        float sz = sin(euler.z);

        mat3 mat;
        mat[0][0] = cy * cz;
        mat[0][1] = cy * sz;
        mat[0][2] = -sy;

        mat[1][0] = sy * sx * cz - cx * sz;
        mat[1][1] = sy * sx * sz + cx * cz;
        mat[1][2] = cy * sx;

        mat[2][0] = sy * cx * cz + sx * sz;
        mat[2][1] = sy * cx * sz - sx * cz;
        mat[2][2] = cy * cx;
        return mat;
    }""",
}.toTable
