# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.



import ../types
import vmath except Quat
import std/strutils
type Window* = ref object of RootObj

const javaClassName {.strdefine.} = ""
when defined(android) and '.' notin javaClassName:
    {.fatal:"You must provide a fully qualified Java class name, e.g. -d:javaClassName=\"com.example.myapp.MyClass\"".}

const jni_prefix = "Java_" & javaClassName.replace('.','_') & "_"

# import jnim
{.pragma: jni, cdecl, gcsafe.}


{.emit:"void *androidActivity,*androidContext,*globalEnvShit,*globalClasShit;".}
var jniEnv* {.importc:"globalEnvShit",nodecl.}: pointer
var jniClass* {.importc:"globalClasShit",nodecl.}: pointer
var androidContext* {.importc,nodecl.}: pointer
var androidActivity* {.importc,nodecl.}: pointer
var assetManager*: pointer

# var context: JVMObject


import std/bitops
import std/math
import std/sequtils
import std/strformat
import std/strutils
import ./gl
import ../screen
import ../util
import ../graphics/render
import ../input

var global_screen: Screen
var engine: MyouEngine
var main_loop: proc(self: MyouEngine)

proc screen*(window: Window): Screen {.inline.} =
    cast[Screen](window)

proc `screen=`*(window: Window, screen1: Screen) {.inline.} =
    global_screen = screen1

var java_vm: pointer

proc JNI_OnLoad(vm, p: pointer): int32 {.exportc,cdecl,dynlib,stackTrace:off.} =
    java_vm = vm;
    const JNI_VERSION_1_6 = 0x00010006
    return JNI_VERSION_1_6



proc NimMain() {.importc.}
proc createNativeApplication(env, class, asset_manager1: pointer): int64 {.exportc:jni_prefix & "createNativeApplication",cdecl,dynlib,stackTrace:off.} =
    debugEcho "createNativeApplication "
    assetManager = asset_manager1
    try:
        NimMain()
        echo "nimmain end"
    except Exception as e:
        # TODO: use logging
        for line in e.getStackTrace.split '\n':
            debugEcho line
        debugEcho getCurrentExceptionMsg()
    1

proc destroyNativeApplication(env, class: pointer, p: int64) {.exportc:jni_prefix & "destroyNativeApplication",cdecl,dynlib.} =
    debugEcho "destroyNativeApplication " & cast[int](env).tohex
    discard

proc onPause(env, class: pointer, p: int64) {.exportc:jni_prefix & "onPause",cdecl,dynlib.} =
    debugEcho "onPause " & cast[int](env).tohex
    discard

var customOnResume*: proc()

proc onResume(env, class: pointer, p: int64, c,a: pointer) {.exportc:jni_prefix & "onResume",cdecl,dynlib.} =
    debugEcho "onResume " & cast[int](env).tohex
    jniEnv = env; jniClass = class
    androidContext = c
    androidActivity = a
    # context = newJVMObject(cast[jobject](androidContext))
    echo "androidContext ", cast[int](androidContext).tohex
    echo "androidActivity ", cast[int](androidActivity).tohex
    if customOnResume != nil:
        customOnResume()
    discard

proc onGlSurfaceCreated(env, class, p: pointer) {.exportc:jni_prefix & "onGlSurfaceCreated",cdecl,dynlib} =
    debugEcho "onGlSurfaceCreated"
    jniEnv = env; jniClass = class
    engine.renderer.initialize()

proc onDisplayGeometryChanged(env, class: pointer, p: int64, r,w,h: int32) {.exportc:jni_prefix & "onDisplayGeometryChanged",cdecl,dynlib.} =
    debugEcho "onDisplayGeometryChanged " & cast[int](env).tohex
    if global_screen != nil:
        global_screen.resize(w,h,r.int8)



proc make_window*(width, height: int32, title: string): Window =
    nil

proc set_vsync*(window: Window, vsync: bool) =
    discard

var max_messages = 0


proc init_graphics*(engine1: MyouEngine, width, height: int32, title: string,
        opengl_version = 330,
        opengl_es = false,
    ) =

    engine = engine1

    let major = opengl_version div 100
    let minor = opengl_version mod 100 div 10
    let rev = opengl_version mod 10
    assert major >= 3

    # TODO: check GL context is at least the same version
    
    engine.renderer.enqueue proc()=
        when not defined(release) and not defined(emscripten):
            proc f(source: GLenum, etype: GLenum, id: GLuint, severity: GLenum, length: GLsizei, message: cstring, userParam: pointer) {.stdcall.} =
                if id == 131185:
                    # buffer usage hints
                    return
                # if message == "GL_INVALID_OPERATION error generated. Target buffer must be bound.":
                #     return
                if max_messages == 0:
                    return
                # dump (source, etype, id, severity, length)
                echo "OpenGL error: ", message
                max_messages -= 1
                if max_messages == 0:
                    echo "No more OpenGL messages will be shown"
            
            glEnable(GL_DEBUG_OUTPUT)
            glDebugMessageCallback cast[GLdebugProc](f), nil
            glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS)

            max_messages = 100
        
        # glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS)
        # glHint(GL_GENERATE_MIPMAP_HINT, GL_FASTEST)
        # glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST)



proc glfm_breakpoint() {.exportc.} =
    # Add "b glfm_breakpoint" to your debugger startup commands
    discard

proc start_platform_main_loop*(engine1: MyouEngine, main_loop1: proc(self: MyouEngine)) =
    engine = engine1
    main_loop = main_loop1

proc onGlSurfaceDrawFrame(env, class: pointer, p: int64, a,b:bool) {.exportc:jni_prefix & "onGlSurfaceDrawFrame",cdecl,dynlib.} =
    debugEcho "onGlSurfaceDrawFrame " & cast[int](env).tohex
# proc onFrame(env, class: pointer) {.exportc:jni_prefix & "onFrame",cdecl,dynlib.} =
    jniEnv = env; jniClass = class
    if engine == nil:
        return
    try:
        engine.main_loop()
    except Exception as e:
        # TODO: use logging
        for line in e.getStackTrace.split '\n':
            echo line
        echo getCurrentExceptionMsg()
        glfm_breakpoint()
    
    
proc onTouched(env, class: pointer, p: int64, x,y: float32):bool {.exportc:jni_prefix & "onTouched",cdecl,dynlib.} =
    debugEcho "onTouched " & cast[int](env).tohex
    discard
proc hasDetectedPlanes(env, class: pointer, p: int64):bool {.exportc:jni_prefix & "hasDetectedPlanes",cdecl,dynlib.} =
    debugEcho "hasDetectedPlanes " & cast[int](env).tohex
    discard
proc isDepthSupported(env, class: pointer, p: int64):bool {.exportc:jni_prefix & "isDepthSupported",cdecl,dynlib.} =
    debugEcho "isDepthSupported " & cast[int](env).tohex
    discard
proc onSettingsChange(env, class: pointer, p: int64, b: bool) {.exportc:jni_prefix & "onSettingsChange",cdecl,dynlib.} =
    debugEcho "onSettingsChange " & cast[int](env).tohex
    discard

type
    JNINativeInterface* = object
        # WARNING: The fields should be defined in exact same order as they are
        # defined in jni.h to preserve ABI compatibility.

        reserved0: pointer
        reserved1: pointer
        reserved2: pointer
        reserved3: pointer

        when defined(not_TARGET_RT_MAC_CFM_and_ppc): # No idea what this means.
            cfm_vectors: array[225, pointer]

        GetVersion*: proc(env: JNIEnvPtr): int32 {.jni.}

        DefineClass*: proc(env: JNIEnvPtr, name: cstring, loader: pointer, buf: ptr byte, len: int): pointer {.jni.}
        FindClass*: proc(env: JNIEnvPtr, name: cstring): pointer {.jni.}
    
    JNIEnv* = ptr JNINativeInterface
    JNIEnvPtr* = ptr JNIEnv

proc GetJniEnv(): pointer {.exportc,jni,stackTrace:off.} =
    # initJNIThread()
    return jniEnv

proc FindClass(name: cstring): pointer {.exportc,jni,stackTrace:off.} =
    let env = cast[JNIEnvPtr](GetJniEnv())
    return env.FindClass(env, name)

